#!/usr/contrib/bin/perl
# This version of makemap.pl makes a 1024x1024 map with all interesting
# elements (including soldier spawn points, runway resupply, etc).
#
# This script decodes a set of Battlefield 1942 battle ("level") .con
# files and shows you where objects are located. I use it to make maps
# of levels and find out exactly where the ammo boxes are, etc.
#
# You'll need Perl to run the makemap.pl program. Just get it for free
# from Activestate, at
# http://www.activestate.com/Products/Download/Register.plex?id=ActivePerl
#
# To use this program, you unpack a level, e.g. Midway, then go to the Midway
# directory and run it:
#
#     cd (...)/levels/Midway
#     perl (path_to_this_perl_file/)makemap_full.pl
#
# Where stuff is located goes to standard out, the map goes to file Midway2.ppm.
#
# You can also direct the output to a file, e.g.
#
#    perl makemap_full.pl > mapdata.txt
#
# You can also pick just a few .con files to use as input.
# To use this feature, type something like:
#     perl makemap.pl conquest/SoldierSpawns.con StaticObjects.con conquest/ControlPoints.con 
# and the locations of the control points, static objects, and moving
# objects will then be shown. By choosing the order of the files, you choose
# the draw order - the last file is drawn last, on top of the other icons.
#
# For how the map coordinates are computed, download the Mod Development Toolkit from
# http://www.planetbattlefield.com/mdt/ and look for the "Map Coordinates Tutorial".
# 
# Lots of options follow: by changing or commenting out lines of code at the top
# of this file you can make the output map larger or show only certain types of
# objects, have icons blend differently, etc.

# Current version: November 8, 2003

# Change Log:
# Sept. 28, 2003: initial MDT release
# Nov. 8, 2003: Added version 1.5 information, namely the patrol boats.

#---------------------------------------------------------

# USER OPTIONS - feel free to change these.

# Note: you can add new icons, and add new names to search for in the .con map files.
# See DOINIT for how to add an icon, and OUTPUTRESULTS for how to associate text with an icon.

$mapscale = 2;		# set to 2 for map twice as large, 3 for map etc...

$battletype = "conquest" ;	# or "ctf", "objectivemode", "singleplayer", "tdm"

$showammo = 1 ;		# if set to 1, show ammo supply points
$showairammo = 1;	# airplane ammo
$showmedical = 1 ;	# medical supply points
$showrepair = 1 ;	# repair pads
$showhangar = 1 ;	# hangars
$showcontrol = 1;	# bases
$showspawn = 1;		# spawn points
$showemplacements = 1;	# defguns, AA, MGs
$showair = 1;		# planes
$showland = 1;		# land vehicles
$showsea = 1;		# sea vehicles
$showobjective = 1;	# objectives

$overlap_test = 1;	# set to 1 to move icons that overlap, 0 to not move icons at all (1 is recommended)
$overlap_radius = 6;	# if an icon is within this radius of another icon, it gets moved outwards

$makemap = 1 ;	# if set to 0, then no PPM format map will be made
$verbose = 1 ;	# if set to 0, then no spawn point information will be output
$importmap = 1 ;	# if set to 1, reads in the map of the level; if 0, the data is put on a gray background
$backgroundpixel = "90 90 100" ;	# background pixel color, used if the map is not imported.

$mapfile = '';	# 'map.ppm' ;	# output map name - if none set, use the current directory name

$blend = 1 ;	# if set to 1, when a new icon is added, blend it with the ones below it. 0 means overwrite, no blend.
$blendover = 0.7 ;	# how much to put the new icon over the old, i.e. 0.6 means 60% new and 40% old

#------------------------------------------------------------

# Internal stuff - DON'T TOUCH UNLESS YOU KNOW WHAT YOU'RE DOING

# Overview:
# If you look at the script, what it does is reads the AbsolutePosition
# of the various objects from conquest/SoldierSpawns.con, StaticObjects.con,
# conquest/ObjectSpawns.con, and conquest/ControlPoints.con. It then searches
# for what the ObjectTemplate.create name is, like "scoutcar". It then paints
# little icons into this array using the pixel coordinates computed from
# AbsolutePosition. The icons are stored in the perl script itself (easier,
# and you can't lose them this way).
#
# The script also reads in the DDS file textures/ingamemap.dds and puts it
# in an array. Once the icons are all drawn, this DDS data is then transferred
# to the icon map array and the result is out. I do it this way (instead of
# writing the icons onto the map data itself) so that I can avoid blending the
# icons with the map data. That is, I blend each icon with the icon below it
# (so you have a clue when two icons are in the same spot), but just substitute
# the resulting icons onto the map (not blending).
#
# Finally, I output the resulting .PPM file.

$pixels = 512 * $mapscale ;		# output map size

$iconResolution = 7 ;	# icon size
$pixcorner = -int($iconResolution/2);	# number of pixels to upper left corner of icon to be placed

$transparent = "255 192 192" ;	# transparent color in an icon (control points)

# these numbers are from Init.con for the battle, the four numbers after game.setActiveCombatArea
$offsetx = 0 ;	# default is 0
$offsety = 0 ;	# default is 0
$scale = 1024 ;		# third and fourth number always match; from Terrain.con, default is 1024, 2048 for Coral Sea, El Alamein, Bocage, Gazala, Battle Axe, Iwo Jima, Wake, 4096 for Midway, Guadalcanal

$blendunder = 1 - $blendover ;

#----

$| = 1 ;	# turn off output buffering, so we see results if piped, etc
$[ = 1;		# set array base to 1

# a stupid thing: the following will be 1 higher than how many there really are.
$numcontrol = 1;
$numicon = 1;

if ( $importmap ) {
	&READDDS_DXT1();
}
&DOINIT() ;
&PROCESS_ARGS() ;

if ( $makemap ) {
	&OUTPUTMAP();
}
exit 0 ;



sub USAGE {
	print STDERR "Usage: perl makemap.pl [conquest/ControlPoints.con] [StaticObjects.con] [conquest/ObjectSpawns.con]\n\n" ;
	print STDERR "This program shows the map locations of objects on a map. Unpack the map's\n";
	print STDERR ".RFA file (using RFA Extractor 1.1) and go to the bf1942/levels/BattleName\n";
	print STDERR "directory, then run this program. Always put the Init.con file first, as\n" ;
	print STDERR "this provides the map scale. For example:\n";
	print STDERR "    perl makemap.pl Init.con ControlPoints.con StaticObjects.con conquest/ObjectSpawns.con\n\n" ;
	print STDERR "shows the contents of the three files listed after Init.con.\n\n";
	print STDERR "If no Init.con file is given, the default scale of 1024 and offsets of 0\nare used.\n\n";
}

sub DOINIT {

# march of the icons...

# These icons are all 7x7 pixels, RGB order. To add an icon here:
# 1) Save your icon as a .PPM format file. In "Options", set the output to be ASCII.
# You can use free programs such as Irfanview to convert to ASCII PPM.
# Open this .PPM file in a text editor. You can then grab the RGB values and add
# your own icon below. If you want to make pixels transparent in your icon, set the
# color to 255 192 192 (the value of $transparent). To hook up your new icon, see
# OUTPUTRESULTS.

# To take an existing icon and modify it: if you put this text
#P3
#7 7
#255
# as a header (without the "#") and add in the data below for an icon, you now have
# a .ppm type file and can edit it in Paint Shop Pro or other image editor.

$ammostring = sprintf "%s",<<"EOF";
0 0 192 0 0 192 0 0 192 0 0 192 0 0 192 0 0 192 0 0 192
0 0 192 0 0 192 0 0 192 52 218 52 0 0 192 0 0 192 0 0 192
0 0 192 0 0 192 52 218 52 52 218 52 52 218 52 0 0 192 0 0 192
0 0 192 0 0 192 52 218 52 52 218 52 52 218 52 0 0 192 0 0 192
0 0 192 0 0 192 52 218 52 52 218 52 52 218 52 0 0 192 0 0 192 
0 0 192 52 218 52 52 218 52 52 218 52 52 218 52 52 218 52 0 0 192
0 0 192 0 0 192 0 0 192 0 0 192 0 0 192 0 0 192 0 0 192 
EOF
	&CONVERTTOICON( $ammostring ) ;
	%ammoicon = %icon ;

$medicalstring = sprintf "%s",<<"EOF";
255 255 255 255 255 255 255 0 0 255 0 0 255 0 0 255 255 255 255 255 255 
255 255 255 255 255 255 255 0 0 255 0 0 255 0 0 255 255 255 255 255 255 
255 0 0 255 0 0 255 0 0 255 0 0 255 0 0 255 0 0 255 0 0 
255 0 0 255 0 0 255 0 0 255 0 0 255 0 0 255 0 0 255 0 0
255 0 0 255 0 0 255 0 0 255 0 0 255 0 0 255 0 0 255 0 0
255 255 255 255 255 255 255 0 0 255 0 0 255 0 0 255 255 255 255 255 255
255 255 255 255 255 255 255 0 0 255 0 0 255 0 0 255 255 255 255 255 255 
EOF
	&CONVERTTOICON( $medicalstring ) ;
	%medicalicon = %icon ;

$medicalboxstring = sprintf "%s",<<"EOF";
255 255 255 255 255 255 255 0 0 255 0 0 255 0 0 255 255 255 255 255 255 
255 255 255 255 255 255 255 0 0 255 0 0 255 0 0 255 255 255 255 255 255 
255 0 0 255 0 0 255 0 0 255 0 0 255 0 0 255 0 0 255 0 0 255 0 0 255 0 0 
255 0 0 255 255 255 255 0 0 255 0 0 255 0 0 255 0 0 255 0 0 255 0 0 255 0 0 
255 0 0 255 0 0 255 0 0 255 255 255 255 255 255 255 0 0 255 0 0 255 0 0 
255 255 255 255 255 255 255 255 255 255 255 255 255 0 0 255 0 0 255 0 0 
255 255 255 255 255 255 
EOF
	&CONVERTTOICON( $medicalboxstring ) ;
	%medicalboxicon = %icon ;

$airammostring = sprintf "%s",<<"EOF";
64 246 0 64 246 0 64 246 0 64 246 0 64 246 0 64 246 0 64 246 0 64 246 0 
64 246 0 0 0 0 64 246 0 0 0 0 64 246 0 64 246 0 64 246 0 0 0 0 0 0 0 0 0 0 
0 0 0 0 0 0 64 246 0 64 246 0 64 246 0 0 0 0 64 246 0 0 0 0 64 246 0 64 246 0 
64 246 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 64 246 0 64 246 0 64 246 0 0 0 0 
64 246 0 0 0 0 64 246 0 64 246 0 64 246 0 64 246 0 64 246 0 64 246 0 64 246 0 
64 246 0 64 246 0 
EOF
	&CONVERTTOICON( $airammostring ) ;
	%airammoicon = %icon ;

# wrench facing upper left
$hangarstring = sprintf "%s",<<"EOF";
216 233 23 216 233 23 216 233 23 216 233 23 216 233 23 216 233 23 216 233 23 
216 233 23 216 233 23 0 0 0 0 0 0 216 233 23 216 233 23 216 233 23 216 233 23 
0 0 0 255 255 0 0 0 0 216 233 23 216 233 23 216 233 23 216 233 23 0 0 0 
0 0 0 0 0 0 0 0 0 216 233 23 216 233 23 216 233 23 216 233 23 216 233 23 
0 0 0 0 0 0 0 0 0 216 233 23 216 233 23 216 233 23 216 233 23 216 233 23 
0 0 0 0 0 0 216 233 23 216 233 23 216 233 23 216 233 23 216 233 23 216 233 23 
216 233 23 216 233 23 
EOF
	&CONVERTTOICON( $hangarstring ) ;
	%hangaricon = %icon ;

# wrench facing upper right
$repairstring = sprintf "%s",<<"EOF";
255 255 0 255 255 0 255 255 0 255 255 0 255 255 0 255 255 0 255 255 0 255 255 0 
255 255 0 255 255 0 0 0 0 0 0 0 255 255 0 255 255 0 255 255 0 255 255 0 
255 255 0 0 0 0 255 255 0 0 0 0 255 255 0 255 255 0 255 255 0 0 0 0 0 0 0 
0 0 0 0 0 0 255 255 0 255 255 0 0 0 0 0 0 0 0 0 0 255 255 0 255 255 0 255 255 0 
255 255 0 0 0 0 0 0 0 255 255 0 255 255 0 255 255 0 255 255 0 255 255 0 
255 255 0 255 255 0 255 255 0 255 255 0 255 255 0 255 255 0 
EOF
	&CONVERTTOICON( $repairstring ) ;
	%repairicon = %icon ;

$controlstring = sprintf "%s",<<"EOF";
255 192 192 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 
255 192 192 255 255 255 255 255 255 0 0 0 0 0 0 0 0 0 255 255 255 255 255 255 
255 255 255 0 0 0 255 255 255 255 255 255 255 255 255 0 0 0 255 255 255 
255 255 255 0 0 0 255 255 255 0 0 0 255 255 255 0 0 0 255 255 255 255 255 255 
0 0 0 255 255 255 255 255 255 255 255 255 0 0 0 255 255 255 255 255 255 
255 255 255 0 0 0 0 0 0 0 0 0 255 255 255 255 255 255 255 192 192 255 255 255 
255 255 255 255 255 255 255 255 255 255 255 255 255 192 192 
EOF
	&CONVERTTOICON( $controlstring ) ;
	%controlicon = %icon ;

$alliescontrolstring = sprintf "%s",<<"EOF";
255 192 192 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 
255 192 192 255 255 255 255 255 255 0 0 0 0 0 0 0 0 0 255 255 255 255 255 255 
255 255 255 0 0 0 0 0 255 0 0 255 0 0 255 0 0 0 255 255 255 255 255 255 
0 0 0 0 0 255 0 0 0 0 0 255 0 0 0 255 255 255 255 255 255 0 0 0 0 0 255 
0 0 255 0 0 255 0 0 0 255 255 255 255 255 255 255 255 255 0 0 0 0 0 0 0 0 0 
255 255 255 255 255 255 255 192 192 255 255 255 255 255 255 255 255 255 
255 255 255 255 255 255 255 192 192 
EOF
	&CONVERTTOICON( $alliescontrolstring ) ;
	%alliescontrolicon = %icon ;

$axiscontrolstring = sprintf "%s",<<"EOF";
255 192 192 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 
255 192 192 255 255 255 255 255 255 0 0 0 0 0 0 0 0 0 255 255 255 255 255 255 
255 255 255 0 0 0 255 0 0 255 0 0 255 0 0 0 0 0 255 255 255 255 255 255 
0 0 0 255 0 0 0 0 0 255 0 0 0 0 0 255 255 255 255 255 255 0 0 0 255 0 0 
255 0 0 255 0 0 0 0 0 255 255 255 255 255 255 255 255 255 0 0 0 0 0 0 0 0 0 
255 255 255 255 255 255 255 192 192 255 255 255 255 255 255 255 255 255 
255 255 255 255 255 255 255 192 192 
EOF
	&CONVERTTOICON( $axiscontrolstring ) ;
	%axiscontrolicon = %icon ;

$spawnstring = sprintf "%s",<<"EOF";
255 64 255 255 64 255 255 64 255 255 64 255 255 64 255 255 64 255 255 64 255 
255 64 255 0 0 0 255 64 255 255 64 255 255 64 255 0 0 0 255 64 255 255 64 255 
255 64 255 0 0 0 255 64 255 0 0 0 255 64 255 255 64 255 255 64 255 255 64 255 
255 64 255 0 0 0 255 64 255 255 64 255 255 64 255 255 64 255 255 64 255 
0 0 0 255 64 255 0 0 0 255 64 255 255 64 255 255 64 255 0 0 0 255 64 255 
255 64 255 255 64 255 0 0 0 255 64 255 255 64 255 255 64 255 255 64 255 
255 64 255 255 64 255 255 64 255 255 64 255 
EOF
	&CONVERTTOICON( $spawnstring ) ;
	%spawnicon = %icon ;

$aastring = sprintf "%s",<<"EOF";
0 255 255 0 255 255 0 255 255 0 255 255 0 255 255 0 255 255 0 255 255 0 255 255 
0 255 255 255 121 1 255 121 1 255 121 1 0 255 255 0 255 255 0 255 255 255 121 1 
255 121 1 0 255 255 255 121 1 255 121 1 0 255 255 0 255 255 255 121 1 255 121 1 
255 121 1 255 121 1 255 121 1 0 255 255 0 255 255 255 121 1 0 255 255 0 255 255 
0 255 255 255 121 1 0 255 255 0 255 255 255 121 1 0 255 255 0 255 255 0 255 255 
255 121 1 0 255 255 0 255 255 0 255 255 0 255 255 0 255 255 0 255 255 0 255 255 
0 255 255 
EOF
	&CONVERTTOICON( $aastring ) ;
	%aaicon = %icon ;

$defgunstring = sprintf "%s",<<"EOF";
255 121 1 255 121 1 255 121 1 255 121 1 255 121 1 255 121 1 255 121 1 255 121 1 
0 255 255 0 255 255 255 121 1 0 255 255 0 255 255 255 121 1 255 121 1 0 255 255 
0 255 255 255 121 1 0 255 255 0 255 255 255 121 1 255 121 1 255 121 1 255 121 1 
255 121 1 255 121 1 255 121 1 255 121 1 255 121 1 0 255 255 0 255 255 255 121 1 
0 255 255 0 255 255 255 121 1 255 121 1 0 255 255 0 255 255 255 121 1 0 255 255 
0 255 255 255 121 1 255 121 1 255 121 1 255 121 1 255 121 1 255 121 1 255 121 1 
255 121 1 
EOF
	&CONVERTTOICON( $defgunstring ) ;
	%defgunicon = %icon ;

$mgstring = sprintf "%s",<<"EOF";
0 255 255 0 255 255 0 255 255 0 255 255 0 255 255 0 255 255 0 255 255 0 255 255 
255 121 1 0 255 255 0 255 255 255 121 1 0 255 255 0 255 255 255 121 1 255 121 1 
255 121 1 0 255 255 0 255 255 255 121 1 0 255 255 255 121 1 255 121 1 255 121 1 
255 121 1 255 121 1 255 121 1 255 121 1 255 121 1 255 121 1 255 121 1 0 255 255 
0 255 255 255 121 1 0 255 255 0 255 255 255 121 1 0 255 255 0 255 255 255 121 1 
0 255 255 0 255 255 0 255 255 0 255 255 0 255 255 0 255 255 0 255 255 0 255 255 
0 255 255 
EOF
	&CONVERTTOICON( $mgstring ) ;
	%mgicon = %icon ;

$atgunstring = sprintf "%s",<<"EOF";
0 255 255 0 255 255 0 255 255 0 255 255 0 255 255 0 255 255 0 255 255 0 255 255 
255 121 1 255 121 1 255 121 1 255 121 1 255 121 1 0 255 255 255 121 1 0 255 255 
0 255 255 0 255 255 0 255 255 0 255 255 255 121 1 255 121 1 0 255 255 0 255 255 
255 121 1 0 255 255 0 255 255 255 121 1 255 121 1 0 255 255 0 255 255 0 255 255 
0 255 255 0 255 255 255 121 1 0 255 255 255 121 1 255 121 1 255 121 1 255 121 1 
255 121 1 0 255 255 0 255 255 0 255 255 0 255 255 0 255 255 0 255 255 0 255 255 
0 255 255 
EOF
	&CONVERTTOICON( $atgunstring ) ;
	%atgunicon = %icon ;



$fighterstring = sprintf "%s",<<"EOF";
255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 
255 255 255 255 255 255 255 90 0 255 90 0 255 90 0 255 90 0 255 90 0 255 255 255 
255 255 255 255 90 0 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 
255 255 255 255 90 0 255 90 0 255 90 0 255 90 0 255 255 255 255 255 255 
255 255 255 255 90 0 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 
255 255 255 255 90 0 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 
255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 
255 255 255 
EOF
	&CONVERTTOICON( $fighterstring ) ;
	%fightericon = %icon ;

$divebomberstring = sprintf "%s",<<"EOF";
255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 
255 255 255 255 255 255 255 255 255 255 90 0 255 255 255 255 255 255 255 255 255 
255 255 255 255 255 255 255 255 255 255 90 0 255 255 255 255 255 255 255 255 255 
255 255 255 255 255 255 255 255 255 255 90 0 255 90 0 255 90 0 255 255 255 
255 255 255 255 255 255 255 255 255 255 90 0 255 255 255 255 255 255 255 90 0 
255 255 255 255 255 255 255 255 255 255 90 0 255 90 0 255 90 0 255 255 255 
255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 
255 255 255 255 255 255 
EOF
	&CONVERTTOICON( $divebomberstring ) ;
	%divebombericon = %icon ;

$levelbomberstring = sprintf "%s",<<"EOF";
255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 
255 255 255 255 255 255 255 255 255 255 90 0 255 255 255 255 255 255 255 255 255 
255 255 255 255 255 255 255 255 255 255 90 0 255 255 255 255 255 255 255 255 255 
255 255 255 255 255 255 255 255 255 255 90 0 255 255 255 255 255 255 255 255 255 
255 255 255 255 255 255 255 255 255 255 90 0 255 255 255 255 255 255 255 255 255 
255 255 255 255 255 255 255 255 255 255 90 0 255 90 0 255 90 0 255 90 0 
255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 
255 255 255 255 255 255 
EOF
	&CONVERTTOICON( $levelbomberstring ) ;
	%levelbombericon = %icon ;

$heavybomberstring = sprintf "%s",<<"EOF";
255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 
255 255 255 255 255 255 255 90 0 255 90 0 255 90 0 255 90 0 255 255 255 
255 255 255 255 255 255 255 90 0 255 255 255 255 255 255 255 255 255 255 90 0 
255 255 255 255 255 255 255 90 0 255 90 0 255 90 0 255 90 0 255 255 255 
255 255 255 255 255 255 255 90 0 255 255 255 255 255 255 255 255 255 255 90 0 
255 255 255 255 255 255 255 90 0 255 90 0 255 90 0 255 90 0 255 255 255 
255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 
255 255 255 255 255 255 
EOF
	&CONVERTTOICON( $heavybomberstring ) ;
	%heavybombericon = %icon ;



$jeepstring = sprintf "%s",<<"EOF";
255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 
255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 
255 0 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 
255 0 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 
255 0 255 255 255 255 255 255 255 255 255 255 255 0 255 255 255 255 255 255 255 
255 0 255 255 255 255 255 255 255 255 255 255 255 255 255 255 0 255 255 0 255 
255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 
255 255 255 255 255 255 255 255 255 
EOF
	&CONVERTTOICON( $jeepstring ) ;
	%jeepicon = %icon ;

$halftrackstring = sprintf "%s",<<"EOF";
255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 
255 255 255 255 255 255 255 0 255 255 255 255 255 255 255 255 255 255 255 0 255 
255 255 255 255 255 255 255 0 255 255 255 255 255 255 255 255 255 255 255 0 255 
255 255 255 255 255 255 255 0 255 255 0 255 255 0 255 255 0 255 255 0 255 
255 255 255 255 255 255 255 0 255 255 255 255 255 255 255 255 255 255 255 0 255 
255 255 255 255 255 255 255 0 255 255 255 255 255 255 255 255 255 255 255 0 255 
255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 
255 255 255 255 255 255 
EOF
	&CONVERTTOICON( $halftrackstring ) ;
	%halftrackicon = %icon ;

$artillerystring = sprintf "%s",<<"EOF";
255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 
255 255 255 255 255 255 255 255 255 255 0 255 255 0 255 255 0 255 255 0 255 
255 255 255 255 255 255 255 0 255 255 255 255 255 255 255 255 255 255 255 255 255 
255 255 255 255 255 255 255 255 255 255 0 255 255 0 255 255 0 255 255 255 255 
255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 
255 0 255 255 255 255 255 255 255 255 0 255 255 0 255 255 0 255 255 0 255 
255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 
255 255 255 255 255 255 255 255 255 
EOF
	&CONVERTTOICON( $artillerystring ) ;
	%artilleryicon = %icon ;

$medtankstring = sprintf "%s",<<"EOF";
255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 
255 255 255 255 255 255 255 0 255 255 0 255 255 0 255 255 0 255 255 0 255 
255 255 255 255 255 255 255 255 255 255 255 255 255 0 255 255 255 255 255 255 255 
255 255 255 255 255 255 255 255 255 255 255 255 255 0 255 255 255 255 255 255 255 
255 255 255 255 255 255 255 255 255 255 255 255 255 0 255 255 255 255 255 255 255 
255 255 255 255 255 255 255 255 255 255 255 255 255 0 255 255 255 255 255 255 255 
255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 
255 255 255 255 255 255 
EOF
	&CONVERTTOICON( $medtankstring ) ;
	%medtankicon = %icon ;

$hvytankstring = sprintf "%s",<<"EOF";
255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 
255 255 255 255 255 255 255 255 255 255 255 255 255 0 255 255 255 255 255 255 255 
255 255 255 255 255 255 255 255 255 255 0 255 255 0 255 255 0 255 255 255 255 
255 255 255 255 255 255 255 255 255 255 255 255 255 0 255 255 255 255 255 255 255 
255 255 255 255 255 255 255 255 255 255 255 255 255 0 255 255 255 255 255 255 255 
255 255 255 255 255 255 255 255 255 255 255 255 255 0 255 255 0 255 255 255 255 
255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 
255 255 255 255 255 255 
EOF
	&CONVERTTOICON( $hvytankstring ) ;
	%hvytankicon = %icon ;

$lcvpstring = sprintf "%s",<<"EOF";
255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 
255 255 255 255 255 255 255 255 255 255 255 255 10 114 52 10 114 52 255 255 255 
255 255 255 255 255 255 255 255 255 10 114 52 255 255 255 255 255 255 10 114 52 
255 255 255 255 255 255 255 255 255 10 114 52 255 255 255 255 255 255 255 255 255 
255 255 255 255 255 255 255 255 255 10 114 52 255 255 255 255 255 255 10 114 52 
255 255 255 255 255 255 255 255 255 255 255 255 10 114 52 10 114 52 255 255 255 
255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 
255 255 255 255 255 255 
EOF
	&CONVERTTOICON( $lcvpstring ) ;
	%lcvpicon = %icon ;

$substring = sprintf "%s",<<"EOF";
192 192 0 192 192 0 192 192 0 192 192 0 192 192 0 192 192 0 192 192 0 192 192 0 
192 192 0 0 0 192 0 0 192 0 0 192 0 0 192 192 192 0 192 192 0 0 0 192 192 192 0 
192 192 0 192 192 0 192 192 0 192 192 0 192 192 0 192 192 0 0 0 192 0 0 192 
0 0 192 192 192 0 192 192 0 192 192 0 192 192 0 192 192 0 192 192 0 192 192 0 
0 0 192 192 192 0 192 192 0 0 0 192 0 0 192 0 0 192 0 0 192 192 192 0 192 192 0 
192 192 0 192 192 0 192 192 0 192 192 0 192 192 0 192 192 0 192 192 0 
EOF
	&CONVERTTOICON( $substring ) ;
	%subicon = %icon ;

$destroyerstring = sprintf "%s",<<"EOF";
192 192 0 192 192 0 192 192 0 192 192 0 192 192 0 192 192 0 192 192 0 192 192 0 
0 0 192 0 0 192 0 0 192 0 0 192 192 192 0 192 192 0 192 192 0 0 0 192 192 192 0 
192 192 0 192 192 0 0 0 192 192 192 0 192 192 0 0 0 192 192 192 0 192 192 0 
192 192 0 0 0 192 192 192 0 192 192 0 0 0 192 192 192 0 192 192 0 192 192 0 
0 0 192 192 192 0 192 192 0 0 0 192 0 0 192 0 0 192 0 0 192 192 192 0 192 192 0 
192 192 0 192 192 0 192 192 0 192 192 0 192 192 0 192 192 0 192 192 0 
EOF
	&CONVERTTOICON( $destroyerstring ) ;
	%destroyericon = %icon ;

$battleshipstring = sprintf "%s",<<"EOF";
192 192 0 192 192 0 192 192 0 192 192 0 192 192 0 192 192 0 192 192 0 192 192 0 
0 0 192 0 0 192 0 0 192 0 0 192 192 192 0 192 192 0 192 192 0 0 0 192 192 192 0 
192 192 0 192 192 0 0 0 192 192 192 0 192 192 0 0 0 192 0 0 192 0 0 192 
0 0 192 192 192 0 192 192 0 192 192 0 0 0 192 192 192 0 192 192 0 192 192 0 
0 0 192 192 192 0 192 192 0 0 0 192 0 0 192 0 0 192 0 0 192 192 192 0 192 192 0 
192 192 0 192 192 0 192 192 0 192 192 0 192 192 0 192 192 0 192 192 0 
EOF
	&CONVERTTOICON( $battleshipstring ) ;
	%battleshipicon = %icon ;

$carrierstring = sprintf "%s",<<"EOF";
192 192 0 192 192 0 192 192 0 192 192 0 192 192 0 192 192 0 192 192 0 192 192 0 
192 192 0 0 0 192 0 0 192 0 0 192 192 192 0 192 192 0 192 192 0 0 0 192 
192 192 0 192 192 0 192 192 0 192 192 0 192 192 0 192 192 0 0 0 192 192 192 0 
192 192 0 192 192 0 192 192 0 192 192 0 192 192 0 0 0 192 192 192 0 192 192 0 
192 192 0 192 192 0 192 192 0 192 192 0 192 192 0 0 0 192 0 0 192 0 0 192 
192 192 0 192 192 0 192 192 0 192 192 0 192 192 0 192 192 0 192 192 0 192 192 0 
192 192 0 
EOF
	&CONVERTTOICON( $carrierstring ) ;
	%carriericon = %icon ;


$mysterystring = sprintf "%s",<<"EOF";
255 255 255 255 255 255 192 0 0 192 0 0 192 0 0 255 255 255 255 255 255 
255 255 255 192 0 0 255 255 255 255 255 255 255 255 255 192 0 0 255 255 255 
255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 192 0 0 255 255 255 
255 255 255 255 255 255 255 255 255 192 0 0 192 0 0 255 255 255 255 255 255 
255 255 255 255 255 255 255 255 255 192 0 0 255 255 255 255 255 255 255 255 255 
255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 
255 255 255 255 255 255 255 255 255 255 255 255 192 0 0 255 255 255 255 255 255 
255 255 255 
EOF
	&CONVERTTOICON( $mysterystring ) ;
	%mysteryicon = %icon ;


# Stuff for DC
$missilestring = sprintf "%s",<<"EOF";
255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 
255 255 255 255 255 255 255 0 255 255 255 255 255 255 255 255 255 255 255 0 255 
255 255 255 255 255 255 255 0 255 255 0 255 255 255 255 255 0 255 255 0 255 
255 255 255 255 255 255 255 0 255 255 255 255 255 0 255 255 255 255 255 0 255 
255 255 255 255 255 255 255 0 255 255 255 255 255 255 255 255 255 255 255 0 255 
255 255 255 255 255 255 255 0 255 255 255 255 255 255 255 255 255 255 255 0 255 
255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 
255 255 255 255 255 255 
EOF
	&CONVERTTOICON( $missilestring ) ;
	%missileicon = %icon ;

$mobileaastring = sprintf "%s",<<"EOF";
255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 
255 255 255 255 255 255 255 255 255 255 0 255 255 0 255 255 0 255 255 255 255 
255 255 255 255 255 255 255 0 255 255 255 255 255 255 255 255 255 255 255 0 255 
255 255 255 255 255 255 255 0 255 255 0 255 255 0 255 255 0 255 255 0 255 
255 255 255 255 255 255 255 0 255 255 255 255 255 255 255 255 255 255 255 0 255 
255 255 255 255 255 255 255 0 255 255 255 255 255 255 255 255 255 255 255 0 255 
255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 
255 255 255 255 255 255 
EOF
	&CONVERTTOICON( $mobileaastring ) ;
	%mobileaaicon = %icon ;

$forkliftstring = sprintf "%s",<<"EOF";
255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 
255 255 255 255 255 255 255 0 255 255 0 255 255 0 255 255 0 255 255 0 255 
255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 
255 0 255 255 255 255 255 255 255 255 255 255 255 0 255 255 0 255 255 0 255 
255 0 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 
255 0 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 
255 0 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 
255 255 255 255 255 255 
EOF
	&CONVERTTOICON( $forkliftstring ) ;
	%forklifticon = %icon ;

$uraltruckstring = sprintf "%s",<<"EOF";
255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 
255 255 255 255 255 255 255 0 255 255 255 255 255 255 255 255 255 255 255 0 255 
255 255 255 255 255 255 255 0 255 255 255 255 255 255 255 255 255 255 255 0 255 
255 255 255 255 255 255 255 0 255 255 255 255 255 255 255 255 255 255 255 0 255 
255 255 255 255 255 255 255 0 255 255 255 255 255 255 255 255 255 255 255 0 255 
255 255 255 255 255 255 255 255 255 255 0 255 255 0 255 255 0 255 255 255 255 
255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 
255 255 255 255 255 255 
EOF
	&CONVERTTOICON( $uraltruckstring ) ;
	%uraltruckicon = %icon ;

$rocketpackstring = sprintf "%s",<<"EOF";
255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 
255 255 255 255 255 255 128 0 128 128 0 128 128 0 128 128 0 128 255 255 255 
255 255 255 255 255 255 128 0 128 255 255 255 255 255 255 255 255 255 128 0 128 
255 255 255 255 255 255 128 0 128 128 0 128 128 0 128 128 0 128 255 255 255 
255 255 255 255 255 255 128 0 128 255 255 255 255 255 255 128 0 128 255 255 255 
255 255 255 255 255 255 128 0 128 255 255 255 255 255 255 255 255 255 128 0 128 
255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 
255 255 255 255 255 255 
EOF
	&CONVERTTOICON( $rocketpackstring ) ;
	%rocketpackicon = %icon ;

$wasserfallstring = sprintf "%s",<<"EOF";
0 255 255 0 255 255 0 255 255 0 255 255 0 255 255 0 255 255 0 255 255 0 255 255 
255 123 0 0 255 255 0 255 255 0 255 255 255 123 0 0 255 255 0 255 255 255 123 0 
0 255 255 0 255 255 0 255 255 255 123 0 0 255 255 0 255 255 255 123 0 0 255 255 
255 123 0 0 255 255 255 123 0 0 255 255 0 255 255 255 123 0 255 123 0 0 255 255 
255 123 0 255 123 0 0 255 255 0 255 255 255 123 0 0 255 255 0 255 255 0 255 255 
255 123 0 0 255 255 0 255 255 0 255 255 0 255 255 0 255 255 0 255 255 0 255 255 
0 255 255 
EOF
	&CONVERTTOICON( $wasserfallstring ) ;
	%wasserfallicon = %icon ;

$objectivestring = sprintf "%s",<<"EOF";
255 192 192 255 192 192 255 0 0 255 0 0 255 0 0 255 192 192 255 192 192 
255 192 192 255 0 0 255 0 0 190 190 255 255 0 0 255 0 0 255 192 192 255 0 0 
255 0 0 190 190 255 190 190 255 190 190 255 255 0 0 255 0 0 255 0 0 255 0 0 
190 190 255 190 190 255 190 190 255 255 0 0 255 0 0 255 0 0 255 0 0 190 190 255 
190 190 255 190 190 255 255 0 0 255 0 0 255 192 192 255 0 0 255 0 0 190 190 255 
255 0 0 255 0 0 255 192 192 255 192 192 255 192 192 255 0 0 255 0 0 255 0 0 
255 192 192 255 192 192 
EOF
	&CONVERTTOICON( $objectivestring ) ;
	%objectiveicon = %icon ;

$planeswstring = sprintf "%s",<<"EOF";
255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 
255 255 255 255 255 255 255 90 0 255 90 0 255 90 0 255 90 0 255 255 255 
255 255 255 255 255 255 255 90 0 255 255 255 255 255 255 255 255 255 255 90 0 
255 255 255 255 255 255 255 90 0 255 90 0 255 90 0 255 90 0 255 255 255 
255 255 255 255 255 255 255 90 0 255 255 255 255 255 255 255 255 255 255 255 255 
255 255 255 255 255 255 255 90 0 255 255 255 255 255 255 255 255 255 255 255 255 
255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 255 
255 255 255 255 255 255 
EOF
	&CONVERTTOICON( $planeswstring ) ;
	%planeswicon = %icon ;

}

sub CONVERTTOICON {
	($iconstring,%icon) = @_;
	$iconstring =~ tr/\n/ /;
	@fld = split(' ',$iconstring);
	$cnt = 1 ;
	for ( $y = 0 ; $y < 7 ; $y++ ) {
		for ( $x = 0 ; $x < 7 ; $x++ ) {
			$icon{$x,$y} = $fld[$cnt] .' '. $fld[$cnt+1] .' '. $fld[$cnt+2] ;
			$cnt += 3;
		}
	}
}

sub PROCESS_ARGS {
	if ( $#ARGV < 1 ) {
		# &USAGE() ;
		@ARGV = ( "$battletype/SoldierSpawns.con", "$battletype/ControlPoints.con", "ObjectiveMode/ObjectiveSpawners.con", "$battletype/ObjectSpawns.con", "StaticObjects.con" );
	}
	# read in the default scale
	$input = 'Init/Terrain.con' ;
	printf "\n========\nProcessing file $input:\n\n";
	&READDATA();

	# and the overrides for the scale
	$input = 'Init.con' ;
	printf "\n========\nProcessing file $input:\n\n";
	&READDATA();

	foreach $arg (@ARGV) {
		printf "\n========\nProcessing file $arg:\n\n";
		$input = $arg ;
		&READDATA() ;
	}
}

sub READDATA {
	$controlpointfile = $spawnfile = $objectfile = $objectivefile = $staticfile = 0;
	unless (open(DATAFILE,$input)) {
		# optional objective file - don't get upset if it's missing
		if ( $input ne 'ObjectiveMode/ObjectiveSpawners.con' ) {
			printf STDERR "WARNING: Can't open $input: $!\n\n";
			&USAGE() ;
		}
		return ;
	}
	if ( $input =~ /init.con/i ) {	# case independent search for init.con in file's name
		&READINITCON();
	} elsif ( $input =~ /terrain.con/i ) {	# case independent search for terrain.con in file's name
		&READTERRAINCON();
	} else {
		if ( $input =~ /ControlPoints.con/i ) {
			$controlpointfile = 1;
		}
		if ( $input =~ /SoldierSpawns.con/i ) {
			$spawnfile = 1;
		}
		if ( $input =~ /ObjectSpawns.con/i ) {
			$objectfile = 1;
		}
		if ( $input =~ /ObjectiveSpawners.con/i ) {
			$objectivefile = 1;
		}
		if ( $input =~ /StaticObjects.con/i ) {
			$staticfile = 1;
		}
		&READCONTENTS() ;
		&OUTPUTRESULTS() ;
	}
	close(DATAFILE);
}

sub READINITCON {

	while (<DATAFILE>) {
		chop;       # strip record separator
		@fld = split(' ',$_);	# split x/y/z into the 3 values

		if ( $#fld > 0 ) {

			if ( lc($fld[1]) eq "game.setactivecombatarea" ) {
				$offsetx = $fld[2] ;
				$offsety = $fld[3] ;
				$scale = $fld[4] ;
				printf("Found map scale: $scale\n");
			}
		}
	}
}
sub READTERRAINCON {

	while (<DATAFILE>) {
		chop;       # strip record separator
		@fld = split(' ',$_);	# split x/y/z into the 3 values

		if ( $#fld > 0 ) {

			if ( lc($fld[1]) eq "console.worldsize" ) {
				$scale = $fld[2] ;
				printf("Found terrain map scale: $scale\n");
			}
		}
	}
}
sub READCONTENTS {

	$adjScale = $pixels / $scale ;
	$header = 1 ;
	$rem = 0 ;
	while (<DATAFILE>) {
		chop;       # strip record separator
		@fld = split(' ',$_);

		if ( $#fld > 0 ) {

			if ( lc($fld[1]) eq "beginrem" ) {
				$rem = 1 ;
			} elsif ( lc($fld[1]) eq "endrem" ) {
				$rem = 0 ;
			} elsif ( lc($fld[1]) eq "object.create" && $#fld >= 2 ) {
				if ( $rem != 1 ) {
					$createName = $fld[2] ;
				}
			} elsif ( lc($fld[1]) eq "object.absoluteposition" && $#fld >= 2 ) {
			    if ( $rem == 0 ) {
				@xyz = split('/',$fld[2]);
				$x = ($xyz[1] - $offsetx) * $adjScale + $pixcorner ;
				$y = ( $scale + $offsety - $xyz[3] ) * $adjScale + $pixcorner ;
				$intx = int($x);
				$inty = int($y);
				$fracx = $x - $intx ;
				$fracy = $y - $inty ;
				$roundx = int($x+0.5);
				$roundy = int($y+0.5);
				$ax = int(1 + 8 * $x / $pixels) ;
				$ax =~ tr/1-9/A-I/;
				$ay = int(1 + 8 * $y / $pixels) ;
				# printf "$createName is at $ax$ay ($x,$y)\n" ;
				$kk = "$createName is at $ax$ay, pixel $roundx,$roundy" ;
				$obj{$kk}++ ;
				$roundx -= $pixelcorner;
				$roundy -= $pixelcorner;
				$kl = sprintf "$createName is at $ax$ay, x/y/z ($xyz[1], $xyz[2], $xyz[3]), pixel $roundx,$roundy" ;
				$objfulldata{$kl}++ ;
			    }
			}
		}
	}
}

# Scroll down a bit for instructions and examples of how to add your own icons.
sub OUTPUTRESULTS {
	# output results, if desired
	if ( $verbose ) {
		foreach $kk ( sort (keys %objfulldata) ) {
			if ( $objfulldata{$kk} == 1 ) {
				printf "$kk\n" ;
			} else {
				printf "$kk - WARNING, identical objects in same position\n" ;
			}
		}
	}
	if ( $makemap ) {
		foreach $kk ( sort (keys %obj) ) {
		    @fld = split(' ',$kk);
		    @subfld = split(',',$fld[$#fld]);
		    if ( $spawnfile ) {
			if ( $showspawn ) {	# show any position found, since they're all spawns
				&DRAWICON( $subfld[1], $subfld[2], %spawnicon );
			}
		    }
		    if ( $staticfile ) {
			if ( $showammo && $fld[1] =~ /ammobox/i ) {
				&DRAWICON( $subfld[1], $subfld[2], %ammoicon );
			} elsif ( $showairammo && $fld[1] =~ /airplaneammo/i ) {
				&DRAWICON( $subfld[1], $subfld[2], %airammoicon );
			} elsif ( $showmedical && $fld[1] =~ /medic/i ) {
				&DRAWICON( $subfld[1], $subfld[2], %medicalicon );
			} elsif ( $showmedical && $fld[1] =~ /medbox/i ) {
				&DRAWICON( $subfld[1], $subfld[2], %medicalboxicon );

			} elsif ( $showhangar && $fld[1] =~ /hangar/i ) {	# hangar
				&DRAWICON( $subfld[1], $subfld[2], %hangaricon );
			} elsif ( $showrepair && $fld[1] =~ /landrep1_supply/i ) {	# repair pad
				&DRAWICON( $subfld[1], $subfld[2], %repairicon );
			}
		    }

		    if ( $controlpointfile ) {
			if ( $showcontrol && $fld[1] =~ /alliesbase/i ) {
				&DRAWICON( $subfld[1], $subfld[2], %alliescontrolicon );
			} elsif ( $showcontrol && $fld[1] =~ /allies_base/i ) {	# SW
				&DRAWICON( $subfld[1], $subfld[2], %alliescontrolicon );
			} elsif ( $showcontrol && $fld[1] =~ /axisbase/i ) {
				&DRAWICON( $subfld[1], $subfld[2], %axiscontrolicon );
			} elsif ( $showcontrol && ($controlpointfile==1) ) {
				&DRAWICON( $subfld[1], $subfld[2], %controlicon );
			}
		    }

		    if ( $objectivefile ) {
			if ( $showobjective ) {
				&DRAWICON( $subfld[1], $subfld[2], %objectiveicon );
			}
		    }

		    # This is where to hook up your own icons. Decide what sort of object
		    # it is, emplacement, land, air, sea, and then copy a pair of lines
		    # from the elsif statements below. You then just put whatever string
		    # you want to search for between the //'s, and put the name of your
		    # icon on the next line. Done!
		    if ( $objectfile ) {
			if ( $showemplacements && $fld[1] =~ /aagunspawner/i ) {
				&DRAWICON( $subfld[1], $subfld[2], %aaicon );
			} elsif ( $showemplacements && $fld[1] =~ /defgunspawner/i ) {
				&DRAWICON( $subfld[1], $subfld[2], %defgunicon );
			} elsif ( $showemplacements && $fld[1] =~ /machinegunspawner/i ) {
				&DRAWICON( $subfld[1], $subfld[2], %mgicon );
			} elsif ( $showemplacements && $fld[1] =~ /antitankgunspawner/i ) {
				&DRAWICON( $subfld[1], $subfld[2], %atgunicon );
			
			} elsif ( $showair && $fld[1] =~ /fighter/i ) {
				&DRAWICON( $subfld[1], $subfld[2], %fightericon );
			} elsif ( $showair && $fld[1] =~ /divebomberspawner/i ) {
				&DRAWICON( $subfld[1], $subfld[2], %divebombericon );
			} elsif ( $showair && $fld[1] =~ /torpedobomberspawner/i ) {
				&DRAWICON( $subfld[1], $subfld[2], %divebombericon );
			} elsif ( $showair && $fld[1] =~ /levelbomberspawner/i ) {
				&DRAWICON( $subfld[1], $subfld[2], %levelbombericon );
			} elsif ( $showair && $fld[1] =~ /heavybomberspawner/i ) {
				&DRAWICON( $subfld[1], $subfld[2], %heavybombericon );

			} elsif ( $showland && $fld[1] =~ /scoutcar/i ) {
				&DRAWICON( $subfld[1], $subfld[2], %jeepicon );
			} elsif ( $showland && $fld[1] =~ /willyspawner/i ) {
				&DRAWICON( $subfld[1], $subfld[2], %jeepicon );
			} elsif ( $showland && $fld[1] =~ /kubelwagenspawner/i ) {
				&DRAWICON( $subfld[1], $subfld[2], %jeepicon );
			} elsif ( $showland && $fld[1] =~ /apcspawner/i ) {
				&DRAWICON( $subfld[1], $subfld[2], %halftrackicon );
			} elsif ( $showland && $fld[1] =~ /artilleryspawner/i ) {
				&DRAWICON( $subfld[1], $subfld[2], %artilleryicon );
			} elsif ( $showland && $fld[1] =~ /lighttankspawner/i ) {
				&DRAWICON( $subfld[1], $subfld[2], %medtankicon);
			} elsif ( $showland && $fld[1] =~ /shermanspawner/i ) {
				&DRAWICON( $subfld[1], $subfld[2], %medtankicon);
			} elsif ( $showland && $fld[1] =~ /heavytankspawner/i ) {
				&DRAWICON( $subfld[1], $subfld[2], %hvytankicon);
			} elsif ( $showland && $fld[1] =~ /flakpanzer/i ) {
				&DRAWICON( $subfld[1], $subfld[2], %mobileaaicon );	# for SW
			} elsif ( $showland && $fld[1] =~ /flaktank/i ) {
				&DRAWICON( $subfld[1], $subfld[2], %mobileaaicon );	# for SW
			} elsif ( $showland && $fld[1] =~ /flakonly/i ) {
				&DRAWICON( $subfld[1], $subfld[2], %mobileaaicon );	# for SW
			} elsif ( $showland && $fld[1] =~ /mediumtankspawner/i ) {
				&DRAWICON( $subfld[1], $subfld[2], %medtankicon);	# for SW

			} elsif ( $showsea && $fld[1] =~ /lcvpspawner/i ) {
				&DRAWICON( $subfld[1], $subfld[2], %lcvpicon);
			} elsif ( $showsea && $fld[1] =~ /ptBoat/i ) {
				&DRAWICON( $subfld[1], $subfld[2], %lcvpicon);
			} elsif ( $showsea && $fld[1] =~ /submarinespawner/i ) {
				&DRAWICON( $subfld[1], $subfld[2], %subicon);
			} elsif ( $showsea && $fld[1] =~ /destroyerspawner/i ) {
				&DRAWICON( $subfld[1], $subfld[2], %destroyericon);
			} elsif ( $showsea && $fld[1] =~ /battleshipspawner/i ) {
				&DRAWICON( $subfld[1], $subfld[2], %battleshipicon);
			} elsif ( $showsea && $fld[1] =~ /carrierspawner/i ) {
				&DRAWICON( $subfld[1], $subfld[2], %carriericon);

			# Battle of Bilston mistakenly puts airplane ammo in the Objects.con file:
			} elsif ( $showairammo && $fld[1] =~ /AirplaneAmmo/i ) {
				&DRAWICON( $subfld[1], $subfld[2], %airammoicon );

			# add some entries for DC (Desert Combat total mod) names
			} elsif ( $showemplacements && $fld[1] =~ /mortarspawner/i ) {
				&DRAWICON( $subfld[1], $subfld[2], %atgunicon );

			} elsif ( $showland && $fld[1] =~ /scud/i ) {
				&DRAWICON( $subfld[1], $subfld[2], %missileicon );
			} elsif ( $showland && $fld[1] =~ /Missile/i ) {
				&DRAWICON( $subfld[1], $subfld[2], %missileicon );
			} elsif ( $showland && $fld[1] =~ /mlrsspawner/i ) {
				&DRAWICON( $subfld[1], $subfld[2], %artilleryicon );
			} elsif ( $showland && $fld[1] =~ /Artillery/i ) {
				&DRAWICON( $subfld[1], $subfld[2], %artilleryicon );
			} elsif ( $showland && $fld[1] =~ /Rocket/i ) {
				&DRAWICON( $subfld[1], $subfld[2], %artilleryicon );
			} elsif ( $showland && $fld[1] =~ /Artillery/i ) {
				&DRAWICON( $subfld[1], $subfld[2], %artilleryicon );
			} elsif ( $showland && $fld[1] =~ /forklift/i ) {
				&DRAWICON( $subfld[1], $subfld[2], %forklifticon );
			} elsif ( $showland && $fld[1] =~ /mobileaa/i ) {
				&DRAWICON( $subfld[1], $subfld[2], %mobileaaicon );
			} elsif ( $showland && $fld[1] =~ /ural/i ) {
				&DRAWICON( $subfld[1], $subfld[2], %uraltruckicon );

			} elsif ( $showair && $fld[1] =~ /uh-60/i ) {
				&DRAWICON( $subfld[1], $subfld[2], %heavybombericon );
			} elsif ( $showair && $fld[1] =~ /HeliSpawner/i ) {
				&DRAWICON( $subfld[1], $subfld[2], %fightericon );
			} elsif ( $showair && $fld[1] =~ /Helicoptor_Fight/i ) {	# I can't help it that people can't spell
				&DRAWICON( $subfld[1], $subfld[2], %fightericon );
			} elsif ( $showair && $fld[1] =~ /Helicoptor_Transport/i ) {
				&DRAWICON( $subfld[1], $subfld[2], %divebombericon );


			# add some entries for EOD (Eve of Destruction total mod) names
			} elsif ( $showemplacements && $fld[1] =~ /MG/i ) {
				&DRAWICON( $subfld[1], $subfld[2], %mgicon );
			} elsif ( $showemplacements && $fld[1] =~ /flak/i ) {
				&DRAWICON( $subfld[1], $subfld[2], %aaicon );
			} elsif ( $showemplacements && $fld[1] =~ /AA/ ) {
				&DRAWICON( $subfld[1], $subfld[2], %aaicon );

			} elsif ( $showland && $fld[1] =~ /jeep/i ) {
				&DRAWICON( $subfld[1], $subfld[2], %jeepicon );
			} elsif ( $showland && $fld[1] =~ /GAZ/i ) {
				&DRAWICON( $subfld[1], $subfld[2], %jeepicon );
			} elsif ( $showland && $fld[1] =~ /BTR/i ) {
				&DRAWICON( $subfld[1], $subfld[2], %halftrackicon );
			} elsif ( $showland && $fld[1] =~ /M113/i ) {
				&DRAWICON( $subfld[1], $subfld[2], %halftrackicon );
			} elsif ( $showland && $fld[1] =~ /V100/i ) {
				&DRAWICON( $subfld[1], $subfld[2], %halftrackicon );
			} elsif ( $showland && $fld[1] =~ /APC/i ) {
				&DRAWICON( $subfld[1], $subfld[2], %halftrackicon );
			} elsif ( $showland && $fld[1] =~ /PT76Spawn/i ) {
				&DRAWICON( $subfld[1], $subfld[2], %artilleryicon );
			} elsif ( $showland && $fld[1] =~ /type63/i ) {
				&DRAWICON( $subfld[1], $subfld[2], %mobileaaicon );
			} elsif ( $showland && $fld[1] =~ /tank/i ) {
				&DRAWICON( $subfld[1], $subfld[2], %medtankicon);
			
			} elsif ( $showair && $fld[1] =~ /Fighter1/i ) {
				&DRAWICON( $subfld[1], $subfld[2], %fightericon );
			} elsif ( $showair && $fld[1] =~ /Plane/i ) {
				&DRAWICON( $subfld[1], $subfld[2], %fightericon );
			} elsif ( $showair && $fld[1] =~ /F105/i ) {
				&DRAWICON( $subfld[1], $subfld[2], %fightericon );
			} elsif ( $showair && $fld[1] =~ /MiG/i ) {
				&DRAWICON( $subfld[1], $subfld[2], %fightericon );
			} elsif ( $showair && $fld[1] =~ /Huey/i ) {
				&DRAWICON( $subfld[1], $subfld[2], %divebombericon );
			} elsif ( $showair && $fld[1] =~ /Chinook/i ) {
				&DRAWICON( $subfld[1], $subfld[2], %divebombericon );
			} elsif ( $showair && $fld[1] =~ /Allied_Jet/i ) {
				&DRAWICON( $subfld[1], $subfld[2], %levelbombericon );
			} elsif ( $showair && $fld[1] =~ /bell/i ) {
				&DRAWICON( $subfld[1], $subfld[2], %heavybombericon );

			} elsif ( $showsea && $fld[1] =~ /PibberSpawn/i ) {
				&DRAWICON( $subfld[1], $subfld[2], %subicon);
			} elsif ( $showsea && $fld[1] =~ /PBR/i ) {
				&DRAWICON( $subfld[1], $subfld[2], %subicon);
			} elsif ( $showsea && $fld[1] =~ /Carrier/i ) {
				&DRAWICON( $subfld[1], $subfld[2], %carriericon);


			# Eternal Silence total mod
			} elsif ( $showland && $fld[1] =~ /RazorbackSpawner/i ) {
				&DRAWICON( $subfld[1], $subfld[2], %jeepicon );
			} elsif ( $showair && $fld[1] =~ /EchelonSpawner/i ) {
				&DRAWICON( $subfld[1], $subfld[2], %fightericon );

			# Battle of Britain mod
			} elsif ( $showcontrol && $fld[1] =~ /RadarTower/i ) {
				$controlpointfile = 1 ;	# signal that this is a control point
				&DRAWICON( $subfld[1], $subfld[2], %alliescontrolicon );
				$controlpointfile = 0 ;
			} elsif ( $showcontrol && $fld[1] =~ /FactorySpawner/i ) {
				$controlpointfile = 1 ;	# signal that this is a control point
				&DRAWICON( $subfld[1], $subfld[2], %alliescontrolicon );
				$controlpointfile = 0 ;

			# Secret Weapons mod
			# (note that mediumtankspawner is done above, so it will show up as heavy)
			} elsif ( $showair && $fld[1] =~ /Goblin/i ) {
				# should ignore this one on Hellendoorn - doubles with Natter
				&DRAWICON( $subfld[1], $subfld[2], %planeswicon );
			} elsif ( $showair && $fld[1] =~ /Natter/i ) {
				&DRAWICON( $subfld[1], $subfld[2], %planeswicon );
			} elsif ( $showair && $fld[1] =~ /Horton/i ) {
				&DRAWICON( $subfld[1], $subfld[2], %planeswicon );
			} elsif ( $showair && $fld[1] =~ /Mustang/i ) {
				&DRAWICON( $subfld[1], $subfld[2], %fightericon );
			} elsif ( $showair && $fld[1] =~ /aw52/i ) {
				&DRAWICON( $subfld[1], $subfld[2], %planeswicon );
			} elsif ( $showair && $fld[1] =~ /c47/i ) {
				&DRAWICON( $subfld[1], $subfld[2], %levelbombericon );
			} elsif ( $showland && $fld[1] =~ /Wasserfall/i ) {
				&DRAWICON( $subfld[1], $subfld[2], %wasserfallicon );
			} elsif ( $showland && $fld[1] =~ /lvt4/i ) {
				&DRAWICON( $subfld[1], $subfld[2], %lcvpicon );
			} elsif ( $showland && $fld[1] =~ /Amphibious/i ) {
				&DRAWICON( $subfld[1], $subfld[2], %lcvpicon );
			} elsif ( $showland && $fld[1] =~ /AmphibSpawner/i ) {
				&DRAWICON( $subfld[1], $subfld[2], %lcvpicon );
			} elsif ( $showland && $fld[1] =~ /DingySpawner/i ) {
				&DRAWICON( $subfld[1], $subfld[2], %lcvpicon );
			} elsif ( $showland && $fld[1] =~ /landingcraft/i ) {
				&DRAWICON( $subfld[1], $subfld[2], %lcvpicon );
			} elsif ( $showland && $fld[1] =~ /JetPack/i ) {
				&DRAWICON( $subfld[1], $subfld[2], %rocketpackicon );
			} elsif ( $showland && $fld[1] =~ /Motorcycle/i ) {
				&DRAWICON( $subfld[1], $subfld[2], %missileicon );
			} elsif ( $showland && $fld[1] =~ /ScoutBike/i ) {
				&DRAWICON( $subfld[1], $subfld[2], %missileicon );
			} elsif ( $showemplacements && $fld[1] =~ /pak40/i ) {
				&DRAWICON( $subfld[1], $subfld[2], %atgunicon );
			} elsif ( $showland && $fld[1] =~ /SturmTiger/i ) {
				&DRAWICON( $subfld[1], $subfld[2], %artilleryicon);
			} elsif ( $showland && $fld[1] =~ /TigerOnly/i ) {
				&DRAWICON( $subfld[1], $subfld[2], %hvytankicon);


			# mystery object in object file - add an icon yourself!
			} elsif ( ($objectfile == 1) && ($showemplacements==1) && ($showland==1) && ($showsea==1) && ($showair==1) ) {
				&DRAWICON( $subfld[1], $subfld[2], %mysteryicon);
				printf STDERR "Unknown vehicle/emplacement type $fld[1]\n";
			}
		    }
		}
	}
	undef %obj ;
	undef %objfulldata ;
}

sub DRAWICON {
	($mapx, $mapy, %icon1) = @_;
	# record location of control point
	if ( $overlap_test ) {
		if ( $controlpointfile ) {
			$controlx[$numcontrol] = $mapx ;
			$controly[$numcontrol++] = $mapy ;
		} else {
			$subshift_count = 0 ;
			&SHIFTICON();
		}
		# add icon to stuff that matters only if read in after control point
		if ( $numcontrol ) {
			$iconx[$numicon] = $mapx ;
			$icony[$numicon++] = $mapy ;
		}
	}
	for ( $y = $mapy ; $y < $mapy + $iconResolution ; $y++ ) {
		for ( $x = $mapx ; $x < $mapx + $iconResolution ; $x++ ) {
			$newclr = $icon1{$x-$mapx,$y-$mapy};
			if ( $newclr ne $transparent ) {
				if ( $blend && exists($rgb{$x,$y}) ) {
					# blend pixels
					@destrgb = split(' ',$rgb{$x,$y});
					@srcrgb = split(' ',$newclr );
					$r = int( $blendunder * $destrgb[1] + $blendover * $srcrgb[1] ) ;
					$g = int( $blendunder * $destrgb[2] + $blendover * $srcrgb[2] ) ;
					$b = int( $blendunder * $destrgb[3] + $blendover * $srcrgb[3] ) ;
					$rgb{$x,$y} = $r .' '. $g .' '. $b ;
				} else {
					$rgb{$x,$y} = $newclr ;
				}
			}
		}
	}
}

sub SHIFTICON {
    # we use $mapx and $mapy directly from the calling routine, as we're going to modify these.

    if ( $numcontrol <= 1 ) {
	# don't do anything until there is a control point read in.
	return ;
    }
    if ( $subshift_count == 0 ) {
	$origmapx = $mapx ;
	$origmapy = $mapy ;
	$useorig = 1 ;
    } elsif ( $subshift_count > 100 ) {
	printf "WARNING: FAILED TO SHIFT $mapx,$mapy SUCCESSFULLY\n";
	return ;
    }
    # First, do we overlap anything?
    $subshift_count++;
    $overlap_radius2 = $overlap_radius * $overlap_radius;
    $overlap_occurs = 0 ;
    for ( $i = 1 ; $i < $numicon ; $i++ ) {
	$dx = $iconx[$i] - $mapx ;
	$dy = $icony[$i] - $mapy ;
	$dist = $dx * $dx + $dy * $dy ;
#printf "  for object $i $iconx[$i],$icony[$i] is at distance $dist\n";
	if ( $dist < $overlap_radius2 ) {
#printf ">>> overlap occurs for object $i $iconx[$i],$icony[$i] at distance $dist\n";
	    $overlap_occurs++ ;
    	    if ( $useorig != 2 ) {
		$ovx = $iconx[$i];
		$ovy = $icony[$i];
		$ovdist = $dist ;
    	    }
#printf "$fld[1] found $overlap_occurs overlapper $ovx,$ovy with distance $ovdist for $mapx,$mapy\n";
	}
    }

    if ( $overlap_occurs ) {
	if ( $subshift_count > $overlap_radius || $overlap_occurs > 1 ) {
#printf "$fld[1] $subshift_count double overlap or regress\n";
	    if ( $useorig == 1 ) {
		$mapx = $origmapx ;
		$mapy = $origmapy ;
#printf "$fld[1] use origin getting set - map now $mapx,$mapy\n";
		$useorig = 2;

	    	# move the object away from the nearest control point, since we overlap two or
	    	# more objects, or we are recursively calling and must stop, or the icon is
	    	# fully on top of another icon.
	    	$mindist = 99999999 ;
    	    	for ( $i = 1 ; $i < $numcontrol ; $i++ ) {
			$dx = $controlx[$i] - $mapx ;
			$dy = $controly[$i] - $mapy ;
			$dist = $dx * $dx + $dy * $dy ;
#printf "$fld[1]   control is at $ovx,$ovy with distance $dist\n";
			if ( $dist < $mindist ) {
				$ovx = $controlx[$i];
				$ovy = $controly[$i];
				$mindist = $dist ;
#printf "$fld[1] found control $ovx,$ovy with minimum distance $dist\n";
			}
	    	}
	    	# OK, have the closest control point.
	    	$ovdist = $mindist;
	    }
	}

	# shift; if the two overlap entirely, shift the icon over
	if ( $ovdist == 0 ) {
	    $ovx++ ;
	    $ovdist = 1;
    	}

    	# move the icon with respect to the single overlapped icon.
    	$dx = $mapx - $ovx;
    	$dy = $mapy - $ovy;
    
	$length = sqrt( $dx * $dx + $dy * $dy ) ;
#printf "$fld[1] with overlapper $ovx,$ovy and map $mapx,$mapy we have $length\n";
	$diffx = int( 0.5 + abs($dx) / $length ) * ($dx >= 0 ? 1 : -1);
	$diffy = int( 0.5 + abs($dy) / $length ) * ($dy >= 0 ? 1 : -1);
	$mapx += $diffx;
	$mapy += $diffy;
#printf "$fld[1] shift $mapx,$mapy by %d %d\n",$diffx,$diffy;
    	# test again
    	&SHIFTICON();
    }
}

sub OUTPUTMAP {
	if ( $mapfile eq '' ) {
		$dir = `pwd`;
		chop $dir;
		@flddir = split('/',$dir);
		if ( $mapscale == 1 ) {
			$outputmapfile = $flddir[$#flddir] . ".ppm";
		} else {
			$outputmapfile = $flddir[$#flddir] . "$mapscale.ppm";
		}
	} else {
		$outputmapfile = $mapfile ;
	}
	unless (open(DATAFILE,'>'.$outputmapfile)) {
		printf STDERR "Can't open $outputmapfile: $!\n";
		&USAGE() ;
		exit 1 ;
	}
	printf STDERR "Creating map file $outputmapfile\n";
	select DATAFILE ;

	if ( $importmap ) {
		$inmapscale = $ddswidth / $pixels ;
	}
	
	printf "P3\n$pixels $pixels\n255\n";
	for ( $y = 0 ; $y < $pixels; $y++ ) {
		for ( $x = 0 ; $x < $pixels; $x++ ) {
			if ( exists($rgb{$x,$y}) ) {
				printf "$rgb{$x,$y} ";
			} else {
				if ( $importmap ) {
					printf "%s ", $inmap{int($x*$inmapscale),int($y*$inmapscale)};
				} else {
					printf $backgroundpixel .' ';
				}
			}
		}
		printf "\n";
	}
	close DATAFILE;	
}

sub READDDS_DXT1 {
	$ddsfile = "Textures/InGameMap.dds" ;
	unless (open(DDSFILE,$ddsfile)) {
		printf STDERR "Can't open $ddsfile: $!\n";
		&USAGE() ;
		exit 1 ;
	}
	binmode(DDSFILE);

	read(DDSFILE,$ddshdr,128);
	@dhdr = unpack "C4I20C44",$ddshdr;
	$ddswidth = $dhdr[7];	# bytes 12-15
	$ddsheight = $dhdr[8];	# bytes 16-19
	$dxtversion = $dhdr[28] - 48 ;

	$blocksize = 8 ;

	$row = 0 ;
	$col = 0 ;
	$endloop = 0 ;
	$skipchunk = 0 ;
	$blockcount = 0 ;
	while (!$endloop && read(DDSFILE,$buff,$blocksize)) {
	    $blockcount++ ;
	    if ( $dxtversion != 3 || ( $skipchunk % 2 == 1 ) ) {
		@ddata = unpack "C8", $buff ;
		# first two bytes are first color, next two second color, RGB 565
		&CONVERT_DDS_COLOR( $ddata[1], $ddata[2] ) ;
		$r0 = $red ; $g0 = $green ; $b0 = $blue ;
		&CONVERT_DDS_COLOR( $ddata[3], $ddata[4] ) ;
		$r1 = $red ; $g1 = $green ; $b1 = $blue ;

		if ( $ddata[2] * 256 + $ddata[1] > $ddata[4] * 256 + $ddata[3] ) { 
			# 2 interp
   			$r2 = int( ( 2 * $r0 + $r1 ) / 3 ) ;
   			$g2 = int( ( 2 * $g0 + $g1 ) / 3 ) ;
   			$b2 = int( ( 2 * $b0 + $b1 ) / 3 ) ;

   			$r3 = int( ( 2 * $r1 + $r0 ) / 3 ) ;
   			$g3 = int( ( 2 * $g1 + $g0 ) / 3 ) ;
   			$b3 = int( ( 2 * $b1 + $b0 ) / 3 ) ;
		} else {
			# 1 interp
			$r2 = int( ( $r0 + $r1 ) / 2 ) ;
			$g2 = int( ( $g0 + $g1 ) / 2 ) ;
			$b2 = int( ( $b0 + $b1 ) / 2 ) ;

			# bogus color, reality check
			$r3 = 255 ;
			$g3 = 0 ;
			$b3 = 0 ;
		}

		# we have to start with element 1, unfortunately...
		$drgb[1] = $r0 .' '. $g0 .' '. $b0;
		$drgb[2] = $r1 .' '. $g1 .' '. $b1;
		$drgb[3] = $r2 .' '. $g2 .' '. $b2;
		$drgb[4] = $r3 .' '. $g3 .' '. $b3;
#printf "drgb: %s, %s, %s, %s\n", $drgb[1], $drgb[2], $drgb[3], $drgb[4];

		$y = $row ;
		for ( $byteno = 5 ; $byteno <= 8 ; $byteno++ ) {
			$bytemap = $ddata[$byteno];
			$x = $col ;
			$inmap{$x++,$y} = $drgb[1+($bytemap%4)];
			$inmap{$x++,$y} = $drgb[1+(($bytemap>>2)%4)];
			$inmap{$x++,$y} = $drgb[1+(($bytemap>>4)%4)];
			$inmap{$x++,$y} = $drgb[1+($bytemap>>6)];
			$y++ ;
		}

		# move on to next tile
		$col += 4 ;
		if ( $col >= $ddswidth ) {
			$col = 0 ;
			$row += 4 ;
			if ( $row > $ddsheight ) {
				$endloop = 1 ;	#should end naturally, but just in case...
			}
		}
#printf "$ddata[1], $ddata[2], $ddata[3], $ddata[4], $ddata[5], $ddata[6], $ddata[7], $ddata[8]\n" ;
	    }
# corrective code, but I think it's unnecessary
#	    	elsif ( $dxtversion == 3 ) {
#		# we're skipping, so check if it's FF... data; if not, skip 133 bytes of junk
#		@ddata = unpack "C8", $buff ;
#		if ( $ddata[1] != 255 ) {
#printf STDERR "ddata is $ddata[1] at block $blockcount\n" ;
#printf STDERR "SKIPPING\n";
#			read( DDSFILE, $skipdata, 133 ) ;
#		}
#	    }
	    $skipchunk = !$skipchunk ;
	}		

	# output map to PPM, if you wish
	if ( 0 ) {
		printf "P3\n$ddswidth $ddsheight\n255\n";
		for ( $y = 0 ; $y < $ddsheight ; $y++ ) {
			for ( $x = 0 ; $x < $ddswidth ; $x++ ) {
				printf "$inmap{$x,$y} ";
			}
			printf "\n";
		}
	}
	close DDSFILE;
}

sub CONVERT_DDS_COLOR {
	($by2, $by1) = @_;
	$red = int((255.9999/31) * ($by1 >> 3));
	$green = int((255.9999/63) * ((( $by1 % 8 )<<3) + ($by2>>5)));
	$blue = int((255.9999/31) * ($by2 % 32));
#printf "bytes are $by1, $by2: $red $green $blue\n";
}


sub by_rating { $author_value{$b} <=> $author_value{$a} ; }
sub by_efficiency { $author_efficiency{$b} <=> $author_efficiency{$a} ; }
sub by_value { $tm_value{$b} <=> $tm_value{$a} ; }
sub numerically { $a <=> $b ; }

